@@ -24,7 +24,7 @@
 QTEST_GUILESS_MAIN(RFCCodecsTest)
 
 #include "rfccodecs.h"
-using namespace KIMAP2;
+using namespace KIMAP;
 
 void RFCCodecsTest::testIMAPEncoding()
 {
@@ -69,10 +69,10 @@ void RFCCodecsTest::testIMAPEncoding()
 
 void RFCCodecsTest::testQuotes()
 {
-    QString test(QStringLiteral("tom\"allen"));
+    QString test(QLatin1String("tom\"allen"));
     QCOMPARE(quoteIMAP(test), QString::fromLatin1("tom\\\"allen"));
-    test = QStringLiteral("tom\'allen");
+    test = QLatin1String("tom\'allen");
     QCOMPARE(quoteIMAP(test), QString::fromLatin1("tom\'allen"));
-    test =  QStringLiteral("tom\\allen");
+    test =  QLatin1String("tom\\allen");
     QCOMPARE(quoteIMAP(test), QString::fromLatin1("tom\\\\allen"));
 }
