@@ -19,6 +19,7 @@
 
 #include "getacljob.h"
 
+#include <KLocalizedString>
 #include "kimap_debug.h"
 
 #include "acljobbase_p.h"
@@ -26,7 +27,7 @@
 #include "session_p.h"
 #include "rfccodecs.h"
 
-namespace KIMAP2
+namespace KIMAP
 {
 class GetAclJobPrivate : public AclJobBasePrivate
 {
@@ -38,10 +39,10 @@ public:
 };
 }
 
-using namespace KIMAP2;
+using namespace KIMAP;
 
 GetAclJob::GetAclJob(Session *session)
-    : AclJobBase(*new GetAclJobPrivate(session, "GetAcl"))
+    : AclJobBase(*new GetAclJobPrivate(session, i18n("GetAcl")))
 {
 }
 
@@ -53,13 +54,13 @@ void GetAclJob::doStart()
 {
     Q_D(GetAclJob);
 
-    d->sendCommand("GETACL", '\"' + KIMAP2::encodeImapFolderName(d->mailBox.toUtf8()) + '\"');
+    d->tags << d->sessionInternal()->sendCommand("GETACL", '\"' + KIMAP::encodeImapFolderName(d->mailBox.toUtf8()) + '\"');
 }
 
 void GetAclJob::handleResponse(const Message &response)
 {
     Q_D(GetAclJob);
-//   qCDebug(KIMAP2_LOG) << response.toString();
+//   qCDebug(KIMAP_LOG) << response.toString();
 
     if (handleErrorReplies(response) == NotHandled) {
         if (response.content.size() >= 4 &&
