@@ -1,7 +1,6 @@
 /*
     Copyright (c) 2009 Kevin Ottens <ervin@kde.org>
     Copyright (c) 2009 Andras Mantia <amantia@kde.org>
-    Copyright (c) 2017 Christian Mollekopf <mollekopf@kolabsys.com>
 
     This library is free software; you can redistribute it and/or modify it
     under the terms of the GNU Library General Public License as published by
@@ -21,7 +20,9 @@
 
 #include "loginjob.h"
 
+#include <KLocalizedString>
 #include "kimap_debug.h"
+#include <ktcpsocket.h>
 
 #include "job_p.h"
 #include "message_p.h"
@@ -45,7 +46,7 @@ static const sasl_callback_t callbacks[] = {
     { SASL_CB_LIST_END, Q_NULLPTR, nullptr }
 };
 
-namespace KIMAP2
+namespace KIMAP
 {
 class LoginJobPrivate : public JobPrivate
 {
@@ -57,7 +58,7 @@ public:
         Authenticate
     };
 
-    LoginJobPrivate(LoginJob *job, Session *session, const QString &name) : JobPrivate(session, name), q(job)
+    LoginJobPrivate(LoginJob *job, Session *session, const QString &name) : JobPrivate(session, name), q(job), encryptionMode(LoginJob::Unencrypted), authState(Login), plainLoginDisabled(false)
     {
         conn = Q_NULLPTR;
         client_interact = Q_NULLPTR;
@@ -66,12 +67,9 @@ public:
     bool sasl_interact();
 
     bool startAuthentication();
-    void sendPlainLogin();
     bool answerChallenge(const QByteArray &data);
     void sslResponse(bool response);
     void saveServerGreeting(const Message &response);
-    void login();
-    void retrieveCapabilities();
 
     LoginJob *q;
 
@@ -80,24 +78,22 @@ public:
     QString password;
     QString serverGreeting;
 
-    QSsl::SslProtocol encryptionMode = QSsl::UnknownProtocol;
-    bool startTls = false;
+    LoginJob::EncryptionMode encryptionMode;
     QString authMode;
-    AuthState authState = Login;
+    AuthState authState;
     QStringList capabilities;
-    bool plainLoginDisabled = false;
-    bool connectionIsEncrypted = false;
+    bool plainLoginDisabled;
 
     sasl_conn_t *conn;
     sasl_interact_t *client_interact;
 };
 }
 
-using namespace KIMAP2;
+using namespace KIMAP;
 
 bool LoginJobPrivate::sasl_interact()
 {
-    qCDebug(KIMAP2_LOG) << "sasl_interact";
+    qCDebug(KIMAP_LOG) << "sasl_interact";
     sasl_interact_t *interact = client_interact;
 
     //some mechanisms do not require username && pass, so it doesn't need a popup
@@ -112,22 +108,22 @@ bool LoginJobPrivate::sasl_interact()
 
     interact = client_interact;
     while (interact->id != SASL_CB_LIST_END) {
-        qCDebug(KIMAP2_LOG) << "SASL_INTERACT id:" << interact->id;
+        qCDebug(KIMAP_LOG) << "SASL_INTERACT id:" << interact->id;
         switch (interact->id) {
         case SASL_CB_AUTHNAME:
             if (!authorizationName.isEmpty()) {
-                qCDebug(KIMAP2_LOG) << "SASL_CB_[AUTHNAME]: '" << authorizationName << "'";
+                qCDebug(KIMAP_LOG) << "SASL_CB_[AUTHNAME]: '" << authorizationName << "'";
                 interact->result = strdup(authorizationName.toUtf8());
                 interact->len = strlen((const char *) interact->result);
                 break;
             }
         case SASL_CB_USER:
-            qCDebug(KIMAP2_LOG) << "SASL_CB_[USER|AUTHNAME]: '" << userName << "'";
+            qCDebug(KIMAP_LOG) << "SASL_CB_[USER|AUTHNAME]: '" << userName << "'";
             interact->result = strdup(userName.toUtf8());
             interact->len = strlen((const char *) interact->result);
             break;
         case SASL_CB_PASS:
-            qCDebug(KIMAP2_LOG) << "SASL_CB_PASS: [hidden]";
+            qCDebug(KIMAP_LOG) << "SASL_CB_PASS: [hidden]";
             interact->result = strdup(password.toUtf8());
             interact->len = strlen((const char *) interact->result);
             break;
@@ -138,19 +134,20 @@ bool LoginJobPrivate::sasl_interact()
         }
         interact++;
     }
-    //FIXME This should return false at least in some cases
     return true;
 }
 
 LoginJob::LoginJob(Session *session)
-    : Job(*new LoginJobPrivate(this, session, QString::fromUtf8("Login")))
+    : Job(*new LoginJobPrivate(this, session, i18n("Login")))
 {
-    qCDebug(KIMAP2_LOG) << this;
+    Q_D(LoginJob);
+    connect(d->sessionInternal(), SIGNAL(encryptionNegotiationResult(bool)), this, SLOT(sslResponse(bool)));
+    qCDebug(KIMAP_LOG) << this;
 }
 
 LoginJob::~LoginJob()
 {
-    qCDebug(KIMAP2_LOG) << this;
+    qCDebug(KIMAP_LOG) << this;
 }
 
 QString LoginJob::userName() const
@@ -189,112 +186,86 @@ void LoginJob::setPassword(const QString &password)
     d->password = password;
 }
 
-/*
- * The IMAP authentication procedure is unfortunately ridiculously complicated due to the many different options:
- *
- * An IMAP Session always has the following structure:
- * * Connection is established.
- * * Server sends greeting.
- * * Client authenticates somehow.
- * * .....
- *
- * If the we have a plain connection it's simple:
- * * Wait for the greeting
- * * Login using the chosen authentication mechanism
- *
- * If we're using TLS (without STARTTLS, so directly):
- * * Immediately initiate TLS handshake.
- * * Wait for the greeting
- * * Get CAPABILITIES to figure out which AUTH mechs are supported
- * * Login using the chosen authentication mechanism
- *
- * If we're using TLS with STARTTLS:
- * * Wait for the greeting (on the unencrypted connection)
- * * Send STARTTLS and wait for OK
- * * Initiate TLS handshake
- * * Get CAPABILITIES to figure out which AUTH mechs are supported
- * * Login using the chosen authentication mechanism
- */
 void LoginJob::doStart()
 {
     Q_D(LoginJob);
 
-    qCDebug(KIMAP2_LOG) << "doStart" << this;
-
-    connect(d->sessionInternal(), SIGNAL(encryptionNegotiationResult(bool)), this, SLOT(sslResponse(bool)));
-
-    if (session()->state() == Session::Disconnected) {
-        auto guard = new QObject(this);
-        QObject::connect(session(), &Session::stateChanged, guard, [d, guard](KIMAP2::Session::State newState, KIMAP2::Session::State) {
-            qCDebug(KIMAP2_LOG) << "Session state changed" << newState;
-            d->login();
-            delete guard;
-        });
-        if (!d->startTls && d->encryptionMode != QSsl::UnknownProtocol) {
-            //We have to encrypt for the greeting
-            d->sessionInternal()->startSsl(d->encryptionMode);
-        }
-        //We wait for the server greeting
-        return;
-    } else {
-        qCInfo(KIMAP2_LOG) << "Session is ready, carring on";
-        //The session is ready, we can carry on.
-        d->login();
-    }
-
-}
-
-void LoginJobPrivate::login()
-{
+    qCDebug(KIMAP_LOG) << this;
     // Don't authenticate on a session in the authenticated state
-    if (q->session()->isConnected()) {
-        q->setError(LoginJob::UserDefinedError);
-        q->setErrorText(QString::fromUtf8("IMAP session in the wrong state for authentication"));
-        q->emitResult();
+    if (session()->state() == Session::Authenticated || session()->state() == Session::Selected) {
+        setError(UserDefinedError);
+        setErrorText(i18n("IMAP session in the wrong state for authentication"));
+        emitResult();
         return;
     }
 
-    if (startTls) {
-        //With STARTTLS we have to try to upgrade our connection before the login
-        qCInfo(KIMAP2_LOG) << "Starting with tls";
-        authState = LoginJobPrivate::StartTls;
-        sendCommand("STARTTLS", {});
-        return;
-    } else {
-        //If this is supposed to be unecrypted or already encrypted we can retrieve capabilties. Otherwise we wait for the sslResponse.
-        if (encryptionMode == QSsl::UnknownProtocol || connectionIsEncrypted) {
-            retrieveCapabilities();
+    // Trigger encryption negotiation only if needed
+    EncryptionMode encryptionMode = d->encryptionMode;
+
+    switch (d->sessionInternal()->negotiatedEncryption()) {
+    case KTcpSocket::UnknownSslVersion:
+        break; // Do nothing the encryption mode still needs to be negotiated
+
+    // For the other cases, pretend we're going unencrypted as that's the
+    // encryption mode already set on the session
+    // (so for instance we won't issue another STARTTLS for nothing if that's
+    // not needed)
+    case KTcpSocket::SslV2:
+        if (encryptionMode == SslV2) {
+            encryptionMode = Unencrypted;
+        }
+        break;
+    case KTcpSocket::SslV3:
+        if (encryptionMode == SslV3) {
+            encryptionMode = Unencrypted;
+        }
+        break;
+    case KTcpSocket::TlsV1:
+        if (encryptionMode == TlsV1) {
+            encryptionMode = Unencrypted;
+        }
+        break;
+    case KTcpSocket::AnySslVersion:
+        if (encryptionMode == AnySslVersion) {
+            encryptionMode = Unencrypted;
+        }
+        break;
+    }
+
+    if (encryptionMode == SslV2 ||
+            encryptionMode == SslV3 ||
+            encryptionMode == SslV3_1 ||
+            encryptionMode == AnySslVersion) {
+        KTcpSocket::SslVersion version = KTcpSocket::SslV2;
+        if (encryptionMode == SslV3) {
+            version = KTcpSocket::SslV3;
+        }
+        if (encryptionMode == SslV3_1) {
+            version = KTcpSocket::SslV3_1;
+        }
+        if (encryptionMode == AnySslVersion) {
+            version = KTcpSocket::AnySslVersion;
+        }
+        d->sessionInternal()->startSsl(version);
+
+    } else if (encryptionMode == TlsV1) {
+        d->authState = LoginJobPrivate::StartTls;
+        d->tags << d->sessionInternal()->sendCommand("STARTTLS");
+
+    } else  if (encryptionMode == Unencrypted) {
+        if (d->authMode.isEmpty()) {
+            d->authState = LoginJobPrivate::Login;
+            qCDebug(KIMAP_LOG) << "sending LOGIN";
+            d->tags << d->sessionInternal()->sendCommand("LOGIN",
+                    '"' + quoteIMAP(d->userName).toUtf8() + '"' +
+                    ' ' +
+                    '"' + quoteIMAP(d->password).toUtf8() + '"');
         } else {
-            qCInfo(KIMAP2_LOG) << "Waiting for encryption before retrieveing capabilities.";
+            if (!d->startAuthentication()) {
+                emitResult();
+            }
         }
     }
-
-}
-
-void LoginJobPrivate::sslResponse(bool response)
-{
-    qCDebug(KIMAP2_LOG) << "Got an ssl response " << response;
-    connectionIsEncrypted = response;
-    if (response) {
-        //It's possible that we receive the ssl info before we receive the server greeting.
-        //In that case we're still in the Disconnected state and shouldn't retrieve the capabilities just yet.
-        //We'll try again via login once the state changes.
-        if (m_session->state() != Session::Disconnected) {
-            retrieveCapabilities();
-        }
-    } else {
-        q->setError(LoginFailed);
-        q->setErrorText(QString::fromUtf8("Login failed, TLS negotiation failed."));
-        encryptionMode = QSsl::UnknownProtocol;
-        q->emitResult();
-    }
-}
-
-void LoginJobPrivate::retrieveCapabilities()
-{
-    qCDebug(KIMAP2_LOG) << "Retrieving capabilities.";
-    authState = LoginJobPrivate::Capability;
-    sendCommand("CAPABILITY", {});
 }
 
 void LoginJob::handleResponse(const Message &response)
@@ -306,11 +277,11 @@ void LoginJob::handleResponse(const Message &response)
     }
 
     //set the actual command name for standard responses
-    QString commandName = QStringLiteral("Login");
+    QString commandName = i18n("Login");
     if (d->authState == LoginJobPrivate::Capability) {
-        commandName = QStringLiteral("Capability");
+        commandName = i18n("Capability");
     } else if (d->authState == LoginJobPrivate::StartTls) {
-        commandName = QStringLiteral("StartTls");
+        commandName = i18n("StartTls");
     }
 
     enum ResponseCode {
@@ -324,7 +295,7 @@ void LoginJob::handleResponse(const Message &response)
     QByteArray tag = response.content.first().toString();
     ResponseCode code = OK;
 
-    qCDebug(KIMAP2_LOG) << commandName << tag;
+    qCDebug(KIMAP_LOG) << commandName << tag;
 
     if (tag == "+") {
         code = CONTINUATION;
@@ -355,8 +326,8 @@ void LoginJob::handleResponse(const Message &response)
             sasl_dispose(&d->conn);
         }
 
-        setError(LoginFailed);
-        setErrorText(QString("%1 failed, server replied: %2").arg(commandName).arg(QLatin1String(response.toString().constData())));
+        setError(UserDefinedError);
+        setErrorText(i18n("%1 failed, server replied: %2", commandName, QLatin1String(response.toString().constData())));
         emitResult();
         return;
 
@@ -372,7 +343,7 @@ void LoginJob::handleResponse(const Message &response)
                 }
                 ++p;
             }
-            qCInfo(KIMAP2_LOG) << "Capabilities updated: " << d->capabilities;
+            qCDebug(KIMAP_LOG) << "Capabilities updated: " << d->capabilities;
         }
         break;
 
@@ -409,27 +380,28 @@ void LoginJob::handleResponse(const Message &response)
         break;
 
     case OK:
+
         switch (d->authState) {
         case LoginJobPrivate::StartTls:
-            //Start encryption and wait for sslResponse
-            d->sessionInternal()->startSsl(d->encryptionMode);
+            d->sessionInternal()->startSsl(KTcpSocket::TlsV1);
             break;
+
         case LoginJobPrivate::Capability:
             //cleartext login, if enabled
             if (d->authMode.isEmpty()) {
                 if (d->plainLoginDisabled) {
-                    setError(LoginFailed);
-                    setErrorText(QString("Login failed, plain login is disabled by the server."));
+                    setError(UserDefinedError);
+                    setErrorText(i18n("Login failed, plain login is disabled by the server."));
                     emitResult();
                 } else {
-                    d->sendPlainLogin();
+                    d->authState = LoginJobPrivate::Login;
+                    d->tags << d->sessionInternal()->sendCommand("LOGIN",
+                            '"' + quoteIMAP(d->userName).toUtf8() + '"' +
+                            ' ' +
+                            '"' + quoteIMAP(d->password).toUtf8() + '"');
                 }
             } else {
                 bool authModeSupported = false;
-                //PLAIN is always supported as defined in the standard. We should also get an AUTH= capability, but in case a server doesn't properly announce it we'll just accept it anyways.
-                if (d->authMode == "PLAIN") {
-                    authModeSupported = true;
-                }
                 //find the selected SASL authentication method
                 Q_FOREACH (const QString &capability, d->capabilities) {
                     if (capability.startsWith(QLatin1String("AUTH="))) {
@@ -440,8 +412,8 @@ void LoginJob::handleResponse(const Message &response)
                     }
                 }
                 if (!authModeSupported) {
-                    setError(LoginFailed);
-                    setErrorText(QString("Login failed, authentication mode %1 is not supported by the server.").arg(d->authMode));
+                    setError(UserDefinedError);
+                    setErrorText(i18n("Login failed, authentication mode %1 is not supported by the server.", d->authMode));
                     emitResult();
                 } else if (!d->startAuthentication()) {
                     emitResult(); //problem, we're done
@@ -462,7 +434,7 @@ void LoginJob::handleResponse(const Message &response)
     }
 
     if (code == MALFORMED) {
-        setErrorText(QString("%1 failed, malformed reply from the server.").arg(commandName));
+        setErrorText(i18n("%1 failed, malformed reply from the server.", commandName));
         emitResult();
     }
 }
@@ -471,8 +443,8 @@ bool LoginJobPrivate::startAuthentication()
 {
     //SASL authentication
     if (!initSASL()) {
-        q->setError(LoginFailed);
-        q->setErrorText(QString("Login failed, client cannot initialize the SASL library."));
+        q->setError(LoginJob::UserDefinedError);
+        q->setErrorText(i18n("Login failed, client cannot initialize the SASL library."));
         return false;
     }
 
@@ -483,8 +455,8 @@ bool LoginJobPrivate::startAuthentication()
 
     int result = sasl_client_new("imap", m_session->hostName().toLatin1(), Q_NULLPTR, nullptr, callbacks, 0, &conn);
     if (result != SASL_OK) {
-        qCWarning(KIMAP2_LOG) << "sasl_client_new failed with:" << result;
-        q->setError(LoginFailed);
+        qCDebug(KIMAP_LOG) << "sasl_client_new failed with:" << result;
+        q->setError(LoginJob::UserDefinedError);
         q->setErrorText(QString::fromUtf8(sasl_errdetail(conn)));
         return false;
     }
@@ -495,16 +467,15 @@ bool LoginJobPrivate::startAuthentication()
         if (result == SASL_INTERACT) {
             if (!sasl_interact()) {
                 sasl_dispose(&conn);
-                q->setError(LoginFailed);   //TODO: check up the actual error
-                q->setErrorText(QString("sasl_interact failed"));
+                q->setError(LoginJob::UserDefinedError);   //TODO: check up the actual error
                 return false;
             }
         }
     } while (result == SASL_INTERACT);
 
     if (result != SASL_CONTINUE && result != SASL_OK) {
-        qCWarning(KIMAP2_LOG) << "sasl_client_start failed with:" << result;
-        q->setError(LoginFailed);
+        qCDebug(KIMAP_LOG) << "sasl_client_start failed with:" << result;
+        q->setError(LoginJob::UserDefinedError);
         q->setErrorText(QString::fromUtf8(sasl_errdetail(conn)));
         sasl_dispose(&conn);
         return false;
@@ -514,24 +485,14 @@ bool LoginJobPrivate::startAuthentication()
     QByteArray challenge = tmp.toBase64();
 
     if (challenge.isEmpty()) {
-        sendCommand("AUTHENTICATE", authMode.toLatin1());
+        tags << sessionInternal()->sendCommand("AUTHENTICATE", authMode.toLatin1());
     } else {
-        sendCommand("AUTHENTICATE", authMode.toLatin1() + ' ' + challenge);
+        tags << sessionInternal()->sendCommand("AUTHENTICATE", authMode.toLatin1() + ' ' + challenge);
     }
 
     return true;
 }
 
-void LoginJobPrivate::sendPlainLogin()
-{
-    authState = LoginJobPrivate::Login;
-    qCDebug(KIMAP2_LOG) << "sending LOGIN";
-    sendCommand("LOGIN",
-            '"' + quoteIMAP(userName).toUtf8() + '"' +
-            ' ' +
-            '"' + quoteIMAP(password).toUtf8() + '"');
-}
-
 bool LoginJobPrivate::answerChallenge(const QByteArray &data)
 {
     QByteArray challenge = data;
@@ -546,8 +507,7 @@ bool LoginJobPrivate::answerChallenge(const QByteArray &data)
 
         if (result == SASL_INTERACT) {
             if (!sasl_interact()) {
-                q->setError(LoginFailed);   //TODO: check up the actual error
-                q->setErrorText(QString("sasl_interact failed"));
+                q->setError(LoginJob::UserDefinedError);   //TODO: check up the actual error
                 sasl_dispose(&conn);
                 return false;
             }
@@ -555,8 +515,8 @@ bool LoginJobPrivate::answerChallenge(const QByteArray &data)
     } while (result == SASL_INTERACT);
 
     if (result != SASL_CONTINUE && result != SASL_OK) {
-        qCWarning(KIMAP2_LOG) << "sasl_client_step failed with:" << result;
-        q->setError(LoginFailed);   //TODO: check up the actual error
+        qCDebug(KIMAP_LOG) << "sasl_client_step failed with:" << result;
+        q->setError(LoginJob::UserDefinedError);   //TODO: check up the actual error
         q->setErrorText(QString::fromUtf8(sasl_errdetail(conn)));
         sasl_dispose(&conn);
         return false;
@@ -570,14 +530,26 @@ bool LoginJobPrivate::answerChallenge(const QByteArray &data)
     return true;
 }
 
-void LoginJob::setEncryptionMode(QSsl::SslProtocol mode, bool startTls)
+void LoginJobPrivate::sslResponse(bool response)
+{
+    if (response) {
+        authState = LoginJobPrivate::Capability;
+        tags << sessionInternal()->sendCommand("CAPABILITY");
+    } else {
+        q->setError(LoginJob::UserDefinedError);
+        q->setErrorText(i18n("Login failed, TLS negotiation failed."));
+        encryptionMode = LoginJob::Unencrypted;
+        q->emitResult();
+    }
+}
+
+void LoginJob::setEncryptionMode(EncryptionMode mode)
 {
     Q_D(LoginJob);
     d->encryptionMode = mode;
-    d->startTls = startTls;
 }
 
-QSsl::SslProtocol LoginJob::encryptionMode()
+LoginJob::EncryptionMode LoginJob::encryptionMode()
 {
     Q_D(LoginJob);
     return d->encryptionMode;
@@ -601,7 +573,7 @@ void LoginJob::setAuthenticationMode(AuthenticationMode mode)
         break;
     case Anonymous: d->authMode = QStringLiteral("ANONYMOUS");
         break;
-    case XOAuth2: d->authMode = QStringLiteral("XOAUTH2");
+    case XOAuth2: d->authMode = QLatin1String("XOAUTH2");
         break;
     default:
         d->authMode = QStringLiteral("");
@@ -612,20 +584,21 @@ void LoginJob::connectionLost()
 {
     Q_D(LoginJob);
 
-    qCWarning(KIMAP2_LOG) << "Connection to server lost " << d->m_socketError;
-    if (d->m_socketError == QSslSocket::SslHandshakeFailedError) {
-        setError(SslHandshakeFailed);
-        setErrorText(QString::fromUtf8("SSL handshake failed."));
-        emitResult();
-    } else if (d->m_socketError == QSslSocket::HostNotFoundError) {
-        setError(HostNotFound);
-        setErrorText(QString::fromUtf8("Host not found."));
-        emitResult();
-    } else {
-        setError(CouldNotConnect);
-        setErrorText(QString::fromUtf8("Connection to server lost."));
-        emitResult();
+    //don't emit the result if the connection was lost before getting the tls result, as it can mean
+    //the TLS handshake failed and the socket was reconnected in normal mode
+    if (d->authState != LoginJobPrivate::StartTls) {
+        qCWarning(KIMAP_LOG) << "Connection to server lost " << d->m_socketError;
+        if (d->m_socketError == KTcpSocket::SslHandshakeFailedError) {
+            setError(KJob::UserDefinedError);
+            setErrorText(i18n("SSL handshake failed."));
+            emitResult();
+        } else {
+            setError(ERR_COULD_NOT_CONNECT);
+            setErrorText(i18n("Connection to server lost."));
+            emitResult();
+        }
     }
+
 }
 
 void LoginJobPrivate::saveServerGreeting(const Message &response)
